//MockDisk.java

import java.nio.*;

/**
   This class is used to emulate an imaginary disk drive. The drive consists
   of 128 pages each of 6 bytes in length.

   Page 0 is always used for the start of the directory of heap
   files stored in the drive. Pages in the directory are stored
   in Header Page format. @see Page

   The directory consists of a linked list of header pages for different
   heap files stored in the drive. The last page of the directory is marked
   by the fact that its next link is null (i.e., 0).

   Header pages in the directory are considered to be deleted files
      if there full pages list link is the null pointer.
      To mark a page as deleted one should of course also delete
      all of the pages associated with it (but not the header page itself)
      in the PAM. To create a heap file with no data pages
      the full pages list link should be set to EMPTY_PAGE (128).

   Pages 123 - 127 are used to contain a Page Allocation Map
   (PAM). Data bytes in these pages are used to indicate which
   pages on the drive have been allocated. The four data bytes
   of page 123 indicate the allocation of the first 32 blocks of
   the drive; page 124 the next 32 blocks, etc.
   @see Page

   @author Chris Pollett
   @version 2003/10/6

*/
public class MockDisk
{
   /**
      default constructor. Allocates memory for pages in disk drive.
      creates a buffer manager and initializes it. lastly formats the drive.
    */
   public MockDisk()
   {
      Page[] drive = new Page[LAST_DISK_PAGE + 1];
      for(int i=0; i<=LAST_DISK_PAGE; i++)
      {
         drive[i] = new Page();
      }
      driveBuffer = new BufferManager();
      driveBuffer.create(BUFFER_SIZE, drive);
      format();
   }

   /**
      does the same as the default contructor but allows one
      to specify a BufferManager to use. Useful if want to
      subclass BufferManager to do a customized page replacement policy.

      @param buffer - buffer manager to use in creating this instance
          of the drive
    */
   public MockDisk(BufferManager buffer)
   {
      Page[] drive = new Page[LAST_DISK_PAGE + 1];
      for(int i=0; i<=LAST_DISK_PAGE; i++)
      {
         drive[i] = new Page();
      }
      driveBuffer = buffer;
      driveBuffer.create(BUFFER_SIZE, drive);
      format();
   }

   /**
      clears all pages in the drive. Creates the PAM and directory pages
      and allocates these pages in the PAM.

    */
   public void format()
   {
      Page p;
      byte zero = 0;

      // erase all the pages

      for(int i = 0; i <= LAST_DISK_PAGE ; i++)
      {
         p = driveBuffer.readPage(i);

         for(int j = 0; j < PAGE_SIZE; j++)
         {
            p.setByte(j , zero);
         }
         driveBuffer.setDirty(i);
         driveBuffer.releasePage(i);
      }

      int previousPage = NULL_PAGE;
      int nextPage;


      //set next and previous page links in PAM

      for(int i = PAM_START; i <= PAM_END; i++)
      {
         p = driveBuffer.readPage(i);
         p.setPreviousPage(previousPage);
         nextPage = (i < LAST_DISK_PAGE) ? i + 1 : NULL_PAGE;
         p.setNextPage(nextPage);
         previousPage = i;
         driveBuffer.setDirty(i);
         driveBuffer.releasePage(i);

      }

      //allocate pages in PAM and for start of directory

      for(int i = PAM_START; i <= PAM_END; i++)
      {
         allocatePage(i);
      }

      allocatePage(DIR_START);
   }

   /**
      sets the page allocation bit of the given page in the PAM
      to true. i.e., allocated

      @param pageNumber - number of page to allocate
   */
   public void allocatePage(int pageNumber)
   {
      setPageAllocation(pageNumber, true);
   }

   /**
      sets the page allocation bit of the given page in the PAM
      to false. i.e., not allocated

      @param pageNumber - number of page to be deallocated
    */
   public void deallocatePage(int pageNumber)
   {
      setPageAllocation(pageNumber, false);
   }

   /**
      sets the page allocation bit of the given page in the PAM
      to the supplied value

      @param pageNumber - number of page to be set
      @param allocate - boolean of whether to allocate (true) or
         deallocate (false)
    */
   public void setPageAllocation(int pageNumber, boolean allocate)
   {
      isValidNumber(pageNumber, 0, LAST_DISK_PAGE);

      int pagesPerPAMPage = DATA_SIZE << 3; // DATA_SIZE * 8
      int whichByte = DATA_START + ((pageNumber >> 3) % DATA_SIZE);

      int currentIndex = PAM_START;
      Page currentPage = driveBuffer.readPage(currentIndex);

      for(int i = PAM_START; i < PAM_START + pageNumber/pagesPerPAMPage; i++)
      {
         driveBuffer.releasePage(currentIndex);
         currentIndex = currentPage.getNextPage();
         currentPage = driveBuffer.readPage(currentIndex);
      }


      int whichBit = pageNumber & 7;

      if(allocate)
      {
         currentPage.setByte(whichByte, (byte)(currentPage.getByte(whichByte)
            | (1 << whichBit)));
      }
      else
      {
         currentPage.setByte(whichByte, (byte)(currentPage.getByte(whichByte)
            & ~(1 << whichBit)));
      }

      driveBuffer.setDirty(currentIndex);
      driveBuffer.releasePage(currentIndex);
   }


   /**
      Used to create an empty heap file and add it to the drive directory

      <b> Warning does not check if file already exists!!</b>

      @param name - two byte name of heap file
      @param recordSize
   */
   public void createFile(String name, int recordSize)
      throws BufferOverflowException, IndexOutOfBoundsException
   {
      int length = name.length();

      isValidNumber(length, 0, DATA_SIZE-2);
         //filenames at most DATA_SIZE-1 chars

      isValidNumber(recordSize, 1, DATA_SIZE);

      Page p;
      int j;
      int i;

      boolean notAddedName = true;
      int currentDirIndex;
      int next = DIR_START;

      while(notAddedName)
      {
         currentDirIndex = next;
         p = driveBuffer.readPage(currentDirIndex);
         if(isNullPage(p.getByte(HEADER_FULL_INDEX)))
            /* checks if header page not in
               use: either just allocated or a deleted file.
               If yes then use it to create file.
               If page in use then this byte will either point to
               a list of full data pages in the given heap file
               or will point to EMPTY_PAGE, if the full page list is
               empty.
            */
         {
            notAddedName = false;
            for(i = DATA_START, j=0; i < DATA_START + length; i++, j++)
            {

               p.setByte(i, (byte)name.charAt(j) );
            }
            driveBuffer.setDirty(currentDirIndex);
            p.setByte(HEADER_FULL_INDEX, (byte)EMPTY_PAGE);
            p.setByte(HEADER_AVAILABLE_INDEX, (byte)DIR_START);
               //indicates the size of records in this file
            p.setRecordSize(recordSize);
         }

         next = p.getNextPage();

         // check if need to extend directory for one more file
         if(notAddedName && isNullPage(next))
         {

            appendPage(currentDirIndex, firstFreePage());
            next = p.getNextPage();
         }

         driveBuffer.releasePage(currentDirIndex);
      }

   }

   /**
      Computes the first not allocated page listed in the PAM

      return - page number of a free page
    */
   public int firstFreePage() throws BufferOverflowException
   {
      boolean notFound = true;

      Page p;
      int currentIndex = PAM_START;
      int j;
      int data = 0;
      int cnt = 0;

      while( notFound && cnt <= LAST_DISK_PAGE)
      {
         p=driveBuffer.readPage(currentIndex);

         j = DATA_START-1;

         while(j < PAGE_SIZE - 1 && notFound)
         {
            j++;
            data = p.getByte(j) &255;
            if(data < 255) notFound=false;
            cnt += 8;

         }
         driveBuffer.releasePage(currentIndex);
         currentIndex++;
      }

      if(notFound) throw new BufferOverflowException();
      cnt -= 8;
      j=0;

      while((data | (1<<j)) == data)
      {
         j++;
         cnt++;
      }

      return cnt;

   }



   /**
      Creates a string with details of the current drive state

      @return - string with details
    */
   public String toStringDrive()
   {
      String out = "";
      out += "Drive's directory:\n";
      out += toStringDirectory();

      out += "Page Allocation Map (PAM) \n";
      out += "Each row gives current allocation of eight pages.\n";
      out += "First row starts with Page 0.\n";

      out += toStringPAM();

      out +="\nData View of the contents of each page:\n";

      out += toStringPages(0, LAST_DISK_PAGE);

      return out;
   }

   /**
      Creates a string with details of the heap files stored
      in the drives directory

      @return - string with details
    */
   public String toStringDirectory()
   {
      boolean notEnd =true;
      int currentDirIndex = DIR_START;
      int next;
      int i = 0;
      Page p;
      String out ="";

      while( notEnd )
      {
         p = driveBuffer.readPage(currentDirIndex);
         next = p.getNextPage();
         notEnd = !isNullPage(next);

         out += "File " + i + ": " + p.toStringHeaderPage() + "\n";

         driveBuffer.releasePage(currentDirIndex);
         currentDirIndex = next;
         i++;
      }
      return out;
   }

   /**
      Creates a string with details of which blocks
      the PAM thinks are allocated and which not

      @return - string with details
    */
   public String toStringPAM()
   {
      Page p;
      int j = 0;
      int currentPAMPage = PAM_START;
      int bytesLeft = PAM_SIZE;
      int nextPage;
      int tmp;
      String out="";

      while(!isNullPage(currentPAMPage))
      {
         p = driveBuffer.readPage(currentPAMPage);

         tmp = (bytesLeft > DATA_SIZE) ? DATA_SIZE : bytesLeft;

         for(int i = DATA_START; i <  DATA_START+tmp; i++)
         {
            out += p.toStringByte(i) + "\n";
         }

         nextPage = p.getNextPage();
         driveBuffer.releasePage(currentPAMPage);

         currentPAMPage = nextPage;
         bytesLeft -= DATA_SIZE;
      }
      return out;
   }

   /**
      Creates a string which lists out the contents of
      a sequence of data pages from the drive

      @return - the string with contents
    */
   public String toStringPages(int low, int high)
   {
      Page p;
      String out = "";

      for(int i = low; i <= high; i++)
      {
         out += "Page " + i + ": ";

         p = driveBuffer.readPage(i);
         out += p.toStringDataPage() +"\n";

         driveBuffer.releasePage(i);
      }

      return out;
   }

   /* END PUBLIC METHODS */

   /**
      Appends the page next to the page node
      in a heap file. It is assumed that
      page next should have no data (newly allocated).
      So its record count is set to zero. Further
      next's next link is set to null

      @param node - page that will append to
      @param next - page that is being appended
    */
   void appendPage(int node, int next)
   {
      isValidNumber(node, 0, LAST_DISK_PAGE);
      isValidNumber(next, 0, LAST_DISK_PAGE);

      allocatePage(next);

      Page p = driveBuffer.readPage(node);
      p.setNextPage(next);
      driveBuffer.setDirty(node);
      driveBuffer.releasePage(node);

      p = driveBuffer.readPage(next);
      p.setPreviousPage(node);
      p.setNextPage(NULL_PAGE);
      p.setNumberOfRecords(0);
      driveBuffer.setDirty(next);
      driveBuffer.releasePage(next);
   }

   /**
      Checks if the given pageNumber is a null page.
      0 is the null page.

      @param pageNumber - page to check if null pointer
      @return - whether or not was zero page
    */
   boolean isNullPage(int pageNumber)
   {
      return (pageNumber == NULL_PAGE);
   }

   /**
    Used with header pages @see Page of heap files. Is used to
    check that the give heap file has an empty full pages list
    but the file is not marked for deletion.

    @param pageNumber - page index to check
    @return - whether page number is the EMPTY_PAGE
    */
   boolean isEmptyPage(int pageNumber)
   {
      return (pageNumber == EMPTY_PAGE);
   }

   /**
      Checks if number is <=high and >= low. If not throws an
      IndexOutofBoundsException. Method says which class it belongs
      to (maybe could have avoided code duplication)

      @param number - number to check
      @param low - low value of range number should be in
      @param high - high value of range number should be in
      @throws IndexOutOfBoundsException - returned if number not in range
    */
   static void isValidNumber(int number, int low, int high)
      throws IndexOutOfBoundsException
   {
      if(number > high || number < low)
         throw new IndexOutOfBoundsException(
            "(MockDisk Index Exception) Number was:" +
             number + ".\n Maximum allowable is: " + high +
            ".\n Minimum allowable is: " + low + ".\n");
   }

   /*
      Some simple tests to illustrate how the drive works.
   */
   public static void main(String args[])
   {
      MockDisk m = new MockDisk();
      m.createFile("hi",1);
      m.createFile("yo",2);
      m.createFile("ho",3);
      m.createFile("h2",4);
      m.createFile("h1",1);
      m.createFile("y2",2);
      m.createFile("h3",3);
      m.createFile("h4",4);
      m.createFile("h5",1);
      m.createFile("y6",2);
      m.createFile("h7",3);
      m.createFile("h8",4);

      System.out.println(m.toStringDrive());

      for(int i=0; i < BUFFER_SIZE; i++)
      {
        System.out.println(m.driveBuffer.toStringBufferPoolPage(i));
      }
   }

   public final static int LAST_DISK_PAGE = Page.MAX_PAGE_INDEX;
      //index of last page stored in drive

   public final static int BUFFER_SIZE = 4;
      // number of pages in the buffer pool the buffer manager has

   public final static int PAGE_SIZE = Page.PAGE_SIZE;
      // size of a disk page in bytes

   public final static int DATA_SIZE = Page.DATA_SIZE;
      // number of bytes of data that can be stored in a disk page

   public final static int DATA_START = PAGE_SIZE - DATA_SIZE ;
      // byte offset to first data byte in a page.

   public final static int PAM_SIZE = (LAST_DISK_PAGE+1)/8;
         //Number of bytes of data needed for PAM

   public final static int PAM_START = LAST_DISK_PAGE-(PAM_SIZE+1)/DATA_SIZE;
         //first page of PAM

   public final static int PAM_END = LAST_DISK_PAGE;
         //Size of PAM in bytes

   public final static int HEADER_FULL_INDEX = PAGE_SIZE-1;
      // index of the byte used in a header page to store a link to the list
      // of full pages in a heap file

   public final static int HEADER_AVAILABLE_INDEX = PAGE_SIZE-2;
   // index of the byte used in a header page to store a link to the list
   // of avilable pages (those that can store more records) in a heap file

   public final static int EMPTY_PAGE = 128;
   // used ot indicate a heap files full page list is empty
   // but the file is not deleted.

   public static int DIR_START = 0;
         //Location of directory in drive

   public static int NULL_PAGE = 0;
         //Next/prev page pointing into Directory
         //indicates no next/prev page.

   BufferManager driveBuffer;
         // buffer used to manage disk I/Os

}

