//Page.java

/**
    This class is used to represent one page of our MockDisk disk drive

    There are two kinds of pages used in our mock disk drive, both of
    which use instances of this class to store: data pages and header pages.
    For more robust purposes might have used subclassing.

    In a data page the bytes are arranged as

      byte#  Use
      0      low order seven bits - next page link
             high order bit is high order bit of record count
      1      low order seven bits - previous page link
             high order bit is low order bit of record count
      2-5    data bytes

      As mentioned above the two high order bits combine together
      to form a two bit number indicating how many records are stored
      in this data page.

      Data pages are used to store the data in a heap file as well as
      being used in the PAM.

    In a header page the bytes are arranged as

      byte#  Use
      0      low order seven bits - next page link
             high order bit is high order bit of record count
      1      low order seven bits - previous page link
             high order bit is low order bit of record count
      2-3    file name
      4      pageNumber of the list of pages with available space
             in this heap file
      5      pageNumber of the list of pages which are full
             in this heap file

      As mentioned above the two high order bits combine together
      to form a two bit number indicating the record size in bytes
      used in this heap file.

      Header pages are used in the drive directory.  If byte
      5 is set to 0 (NULL_PAGE). It indicates this file is deleted.
      If byte 5 is set to 128 (EMPTY_PAGE). It indicates the
      full page list is empty but the file is still in use
      (not deleted).

    @author Chris Pollett
    @version 2003/10/6

*/

public class Page
{
   /**
      Allocates memory for the bytes of this page
   */
   public Page()
   {
      data = new byte[PAGE_SIZE];
   }

   /**
      Copy constructor for this class

      @param p - Page to be copied
   */
   public Page(Page p)
   {
      data = new byte[PAGE_SIZE];
      copy(p);
   }

   /**
      Copies the page p into the current page object

      @param - page to be copied
   */
   public void copy(Page p)
   {
      for(int i=0; i < PAGE_SIZE; i++)
      {
         data[i] = p.data[i];
      }
   }

   /**
      Gives an integer reference to the next page in the file after
      this page.

      @return - index of the next page
   */
   public int getNextPage()
   {
      return data[0] & MAX_PAGE_INDEX;
   }

   /**
      sets the next page link off this page to the indicated value

      @param pageNumber

   */
   public void setNextPage(int pageNumber)
   {
      isValidNumber(pageNumber, 0, MAX_PAGE_INDEX);

      data[0] &=MASK;
      data[0] += (byte)pageNumber;
   }

   /**
      Gives an integer reference to the previous page in the file before
      this page.

      @return - index of the next page
    */
   public int getPreviousPage()
   {
      return data[1] & MAX_PAGE_INDEX;
   }

   /**
      sets the previous page link off this page to the indicated value

      @param pageNumber
    */
   public void setPreviousPage(int pageNumber)
   {
      isValidNumber(pageNumber, 0, MAX_PAGE_INDEX);

      data[1] &=MASK;
      data[1] += (byte)pageNumber;
   }



   /**
      If the page is a data page then this returns the number
      of records stored in the data page.

      @return - number of records in this page
   */
   public int getNumberOfRecords()
   {
      return  (((data[0] >> 7)& 1) << 1) | (data[1]>>7 &1);
   }

   /**
      If the page is a header page then this returns the number
      of bytes ina record that appears in this heap file.

      @return - number of bytes in a record
    */
   public int getRecordSize()
   {
      return getNumberOfRecords() + 1;
   }

   /**
      Used in a header page of a heap file to indicate the size of
      records stored in the file.

      @param recordSize - size of the records in this heap file in bytes
    */
   public void setRecordSize(int recordSize)
   {
      --recordSize;
      setNumberOfRecords(recordSize);
   }
   /**
      Used in a data page of a heap file to set the value of the number of
      records currently stored in the page.

      @param number - what value to set it to
    */
   public void setNumberOfRecords(int number)
   {
      isValidNumber(number, 0, DATA_SIZE -1);
      data[0] &= 127;
      data[1] &= 127;
      data[0] |= (number & 2)<< 6;
      data[1] |= ( number & 1) << 7;
   }

   /**
      Returns the given index byte from the bytes stored on page

      @param byteNumber - index of byte to return

      @return - a byte from page
    */
   public byte getByte(int byteNumber)
   {
      isValidNumber(byteNumber, 0, PAGE_SIZE-1);
      return data[byteNumber];
   }

   /**
      Sets the value of the specified byte of this page

      @param byteNumber - index of byte to set
      @param value - value to set that byte to
    */
   public void setByte(int byteNumber, byte value)
   {
      isValidNumber(byteNumber, 0, PAGE_SIZE-1);
      data[byteNumber] = value;
   }


   /**
      Generates from byte i of the bytes in this page a
      String of representating the settings of the indiviual bits
      in this byte.

      @return string representation of the requested byte
    */
   public String toStringByte(int i)
   {
      isValidNumber(i, 0, PAGE_SIZE-1);
      int b = getByte(i);
      String out = "";

      for(int j = 1; j <= 128; j <<= 1)
      {
         if( (b & j) > 0)
         {
            out += " 1";
         }
         else
         {
            out += " 0";
         }
      }
      return out;
   }

   /**
      Generates a string that gives informations about the contents
      of the page assuming the page is the data page of some heap file

      @return  - string with information
    */
   public String toStringDataPage()
   {
      String out ="";
      out += "Next Page: "+ getNextPage();
      out += " Previous Page: " + getPreviousPage();
      out += " Record Cnt: " + getNumberOfRecords();
      out += " Data: [";

      for(int j = DATA_START; j < PAGE_SIZE; j++)
      {
         int b = getByte(j) & 255;
         out += " "+ b;
      }
      out += " ]";
      return out;
   }
   /**
      Generates a string that gives informations about the contents
      of the page assuming the page is the header page of some heap file

      @return  - string with information
   */
   public String toStringHeaderPage()
   {
      String out = "Name: ";

      for(int i = DATA_START ; i < PAGE_SIZE - 2; i++)
      {
         out += (char)getByte(i);
      }
      out += " Next Page: "+ getNextPage();
      out += " Previous Page: " + getPreviousPage();

      out += " Record Size: " + getRecordSize();
      out += " Available List Link:" + (getByte(PAGE_SIZE - 2) & 255);
      out += " Full List Link:" + (getByte(PAGE_SIZE - 1) & 255);

      return out;
   }

   /**
      Checks if number is <=high and >= low. If not throws an
      IndexOutofBoundsException. Method says which class it belongs
      to (maybe could have avoided code duplication)

      @param number - number to check
      @param low - low value of range number should be in
      @param high - high value of range number should be in

      @throws IndexOutOfBoundsException - returned if number not in range
    */
   static void isValidNumber(int number, int low, int high)
      throws IndexOutOfBoundsException
   {
      if(number > high || number < low)
         throw new IndexOutOfBoundsException("(Page Index Exception) Number was:"
            + number +".\n Maximum allowable is: " + high +
            ".\n Minimum allowable is: " + low + ".\n");
   }
   public static int PAGE_SIZE=6; //number of bytes in a page

   public static int DATA_SIZE=4; // number of bytes of data in a page
                                  // (ignoring next and previous links)

   public static int DATA_START = PAGE_SIZE - DATA_SIZE ;
                                 //byte offset to first byte of data

   public static int MAX_PAGE_INDEX=127;
                                 // maximum number that can occur
                                 // in a next and previous link

   public static int MASK=128; // used to select for the high order bit of a byte;

   byte[] data; // actual data stored in page

}

